/////////////////////////////////////////////////////////////////////////////
///
/// @file limits2d.h
///
/// Limits set on computations.
/// This file contains the definitions of a few limits imposed on selected
/// aspect of the computations, like the maximal acceptable size
/// of an index pair.
/// These parameters impact the memory consumption of the program
/// so these limits must be increased with care if necessary.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2008 by Pawel Pilarczyk.
//
// This file is part of my research software package.  This is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started on February 9, 2008. Last revision: April 14, 2008.


#ifndef _CMGRAPHS_LIMITS2D_H_
#define _CMGRAPHS_LIMITS2D_H_


namespace custom {

/// Limitations and bounds for 2-dimensional computations.
/// This namespace contains some constants which define various limitations
/// on the depth and strength of computations. These limits are designed
/// for the 2-dimensional computations conducted for the Leslie population
/// model to make sure that the program does not use too much memory
/// (about 2GB max).
namespace limits2d {


// --------------------------------------------------
// ------------------ refinements -------------------
// --------------------------------------------------

/// The number of refinements that should be done if a Morse set
/// with the trivial index is encountered or if the image of a Morse set
/// sticks out of the rectangular region to which the computations
/// are restricted. Set to zero for no refinements.
const int refineDepth = 5;

/// The maximal allowed size of a set of cubes in the phase space
/// which can be refined at the initial subdivision level.
/// Set to zero to suppress any refinements.
const int maxRefineSize0 = 200000;

/// The maximal allowed size of a set of cubes in the phase space
/// which can be refined at the subsequent subdivision levels.
/// Set to zero to suppress any refinements.
const int maxRefineSize1 = 50000;


// --------------------------------------------------
// ------------------- map images -------------------
// --------------------------------------------------

/// The maximal allowed diameter of the cubical image of a signle box.
/// This limit is set up in order to prevent a blow-up of the program
/// in the memory because of accumulating huge images of boxes.
const int maxImageDiameter = 3000;

/// The maximal allowed volume of the cubical image of a single box.
/// This limit is set up in order to prevent a blow-up of the program
/// in the memory because of accumulating huge images of boxes.
const int maxImageVolume = 6000;


// --------------------------------------------------
// ------------------- index pair -------------------
// --------------------------------------------------

/// The maximal allowed size of the index pair. This size is measured
/// as the total number of cubes in the index pair and its forward image.
const int maxIndexPairSize = 1200000;


} // namespace limits2d
} // namespace custom


#endif // _CMGRAPHS_LIMITS2D_H_

